using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;

namespace ReactionDiffusion
{
    public partial class Form1 : Form
    {
        private const int width = 300;
        private const int height = 300;
        private Bitmap bitmap = new Bitmap(width, height, PixelFormat.Format24bppRgb);
        private double[,] a = new double[width, height];
        private double[,] b = new double[width, height];
        private double[,] aAux = new double[width, height];
        private double[,] bAux = new double[width, height];
        private double[,] alpha = new double[width, height];
        private double[,] beta = new double[width, height];

        private Random random = new Random(42);

        public Form1()
        {
            InitializeComponent();

            ClientSize = new Size(width, height);

            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    a[x, y] = 4.0 +0.5 * (random.NextDouble() - 0.5);
                    b[x, y] = 4.0 +0.5 * (random.NextDouble() - 0.5);
                    //15.32 ... 16.2
                    alpha[x, y] = 16.0 + 0.01 * 2.0*(random.NextDouble() - 0.5);
                    //11.7 ... 12.6
                    beta[x, y] = 12.0 + 0.01 * 2.0*(random.NextDouble() - 0.5);
                }
            }
        }

        private void iterate()
        {
            for (int x = 0; x < width; x++)
            {
                int right = (x + 1) % width; // wrap around
                int left = (x - 1 + width) % width;
                for (int y = 0; y < height; y++)
                {
                    int top = (y + 1) % height;
                    int bottom = (y - 1 + height) % height;
                    double a1 = a[x, y] + 0.0005 * (alpha[x, y] - a[x, y] * b[x, y]) + 0.02 * (a[left, y] + a[x, top] + a[right, y] + a[x, bottom] - 4.0 * a[x, y]);
                    double b1 = b[x, y] + 0.0005 * (a[x, y] * b[x, y] - b[x, y] - beta[x, y]) + 0.003 * (b[left, y] + b[x, top] + b[right, y] + b[x, bottom] - 4.0 * b[x, y]);
                    aAux[x, y] = Math.Max(0.01, Math.Min(200.0, a1));
                    bAux[x, y] = Math.Max(0.01, Math.Min(200.0, b1));
                }
            }

            // switch a, aAux and b, bAux
            double[,] c = a;
            a = aAux;
            aAux = c;
            c = b;
            b = bAux;
            bAux = c;
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            for (int i = 0; i < 10; i++)
            {
                iterate();                
            }

            unsafe
            {
                BitmapData bd = bitmap.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb);

                byte* p = (byte*)bd.Scan0;
                for (int y = 0; y < height; y++)
                {
                    byte* q = p;
                    for (int x = 0; x < width; x++)
                    {
                        double red = 30 * (a[x, y] - 0);
                        double green = 30 * (b[x, y] - 0);
                        double blue = 0.0;
                        *q = (byte)Math.Max(0, Math.Min(255, blue));
                        q++;
                        *q = (byte)Math.Max(0, Math.Min(255, green));
                        q++;
                        *q = (byte)Math.Max(0, Math.Min(255, red));
                        q++;
                    }
                    p += bd.Stride;
                }

                bitmap.UnlockBits(bd);
            }

            e.Graphics.DrawImage(bitmap, 0, 0);

            Invalidate();
        }
    }
}